#!/usr/bin/python
# PYTHON Version 2.7
"""How many minimal dominating sets can a tree have? (at most)
"""

states = "OPLcpf"
successor = {
"O":"O--O-O", # new name: *G*ood
"P":"L--P-O", # new name: *S*elf
"L":"L--L-O", # *L*acking
"c":"cc-cc-", # new name: *d*ominated
"p":"---pp-", # *p*rivate
"f":"ccpff-", # *f*ree
}
newnames = "GSLdpf"

def combine(v1,v2):
    """combine two 6-vectors v1 and v2 with the star operation"""
    newnum = dict((s,0) for s in states)
    for k1,s in enumerate(states):
       for k2,t in enumerate(states):
          suc_state = successor[s][k2]
          if suc_state!="-":
              newnum[suc_state]+= v1[k1]*v2[k2]
    return tuple(newnum[s] for s in states)


lambda_inverse = 95 ** (-1./13)
maxpower = 0
def power(i):
    " compute 1/lambda**i "
    global maxpower
    maxpower = max(maxpower,i)
    if i<0: error(0,0)
    s = 1
    for _ in range(i): s*=lambda_inverse
    return s

if abs(power(13) - 1./95) > 1e-14:
    print "Error: lambda_inverse too imprecise"
    exit(1)



###############################
"""
(s,(v1,v2,v3,v4,v5,v6)) denotes
(v1*f**s, v2*f**s, ..., v6*f**s),
where f=lambda_inverse is the factor 1/95**(1/13)
"""

################################

from lambdas import lambdas
from hullvertices import hull_vertices
evaluated_hull = [tuple(x*power(s) for x in v) for s,v in hull_vertices]

howobtained = [""]*len(hull_vertices)
fmt = " = v%d*v%d"
#fmt = "=v_{%d}*v_{%d}" # for LaTeX
for (i,j),(s,t) in lambdas.items():
   if s==-1: # s=-1 signals an exact equality
      k = t.keys()[0]
      howobtained[k]+= fmt %(i+1,j+1)

def error(i,j):
    print "error in processing",i,j
    exit(1)

mingap = 1

for i,(n1,v1) in enumerate(hull_vertices):
 s,v = hull_vertices[i]
 print "v%d = %31sf^%d"%(i+1,str(v),s)+ howobtained[i]
 # for LaTeX:
# exp,vec = hull_vertices[i]
# print "$v_{%d}"%(i+1) +howobtained[i] + "$&${} =",vec, (
#     r"\lambda^{%d}"%(-exp) if exp !=0 else "")+ r"$\\"
         
 for j,(n2,v2) in enumerate(hull_vertices):
   """check if the vector v = h[i] o h[j] is indeed dominated by
   a the convex combination of the vectors in hullset given by
   lambdas[i,j]"""

   # special treatment to allow exact multiplication by 0.9 and
   # exact division by lambda**13=95:
   if (v1==(0.9,0,0,0,0,0) and
        (n1+n2, combine((1,0,0,0,0,0),v2)) == (13,(95,0,0,0,0,0))):
       ex_new = (0, (0.9,0,0,0,0,0))
   else:
       newvec_x = combine(v1,v2)
       ex_new = (n1+n2,newvec_x)
   u,lambo = lambdas[i,j]
   if u==-1: # flag u=-1 indicates exact result == hull_vertex[k]:
     k=lambo.keys()[0]
     if ex_new != hull_vertices[k]:
         error(i,j)
     continue

   s,v = ex_new
   right = tuple(x*power(s) for x in v)

   # coefficients are coded as integers, denoting multiples of 0.0001  
   if sum(lambo.values()) != 10000:
        print ("sum is not 1.")
        error(i,j)
   for v in lambo.values():
     if v<0 or v!=int(v):
        print "lambda negative or noninteger."
        error(i,j)   
   left = tuple(sum(lambo[j]*evaluated_hull[j][k] for j in lambo)*0.0001
        for k in range(6))
   O,P,L,c,p,f = left
   left2 = O,O+P,O+P+L,c,c+p,f

   O,P,L,c,p,f = right
   right2 = O,O+P,O+P+L,c,c+p,f
   
   if not all(x>=y for x,y in zip(left2,right2)):
       error(i,j)
   gap =  min(x-y for x,y in zip(left2,right2) if (x,y)!=(0,0))
   if gap<mingap:
       smallestgap = i,j,lambo," (These indices are 0-based.)"
       mingap = gap

print
print "The largest occurring power of lambda_inverse was",maxpower
print "The smallest gap except 0-0 was", mingap
print "It occurred for", smallestgap
if mingap< 1e-10:
       print "ERROR: gap too small"
       exit(1)
print "All checks passed OK."              

# minimum gap: 0.0000047262: for showing v51*v41 < v21
